""" Django request reader """

from sap.cf_logging import defaults
from sap.cf_logging.core.request_reader import RequestReader


class DjangoRequestReader(RequestReader):
    """ Read log related properties out of Django request """

    def get_remote_user(self, request):
        return request.META.get('REMOTE_USER') or defaults.UNKNOWN

    def get_protocol(self, request):
        return request.scheme

    def get_content_length(self, request):
        return request.META.get('CONTENT_LENGTH') or defaults.UNKNOWN

    def get_remote_ip(self, request):
        return request.META.get('REMOTE_ADDR')

    def get_remote_port(self, request):
        return request.META.get('SERVER_PORT') or defaults.UNKNOWN

    def get_http_header(self, request, header_name, default=None):
        if request is None:
            return default

        if header_name in request.META:
            return request.META.get(header_name)
        if header_name.upper() in request.META:
            return request.META.get(header_name.upper())

        return default



""" Module Context """

_CORRELATION_ID_KEY = 'correlation_id'

class Context(object):  # pylint: disable=useless-object-inheritance
    """ Class for getting and setting context variables """

    def set(self, key, value, request):
        """ Store session variable """
        raise NotImplementedError

    def get(self, key, request):
        """ Get session variable """
        raise NotImplementedError

    def get_correlation_id(self, request=None):
        """ Gets the current correlation_id. Ensure that you are calling this method in the
        appropriate location of your code having in mind which framework you are using. """

        return self.get(_CORRELATION_ID_KEY, request)

    def set_correlation_id(self, value, request=None):
        """ Sets the current correlation_id. Ensure that you are calling this method in the
        appropriate location of your code having in mind which framework you are using. """

        return self.set(_CORRELATION_ID_KEY, value, request)



""" Sanic logging context - used by the logging package to keep
request specific data during request processing for logging purposes.
For example correlation_id needs to be stored so all log entries contain it.
"""
from sap.cf_logging.core.context import Context

CONTEXT_NAME = 'cf_logger_context'


class SanicContext(Context):
    """ Stores logging context in Sanic's request object """

    def set(self, key, value, request):
        if request is not None:
            setattr(request.ctx, key, value)

    def get(self, key, request):
        if request is None:
            return None

        return getattr(request.ctx, key, None)



""" Falcon response reader """
from sap.cf_logging.core.response_reader import ResponseReader

CONTENT_LENGTH = 'Content-Length'


class FalconResponseReader(ResponseReader):
    """ Read log related properties out of falcon response """

    def get_status_code(self, response):
        return response.status.split(' ', 1)[0]

    def get_response_size(self, response):
        return response.get_header('Content-Length')



""" Module for the ResponseReader class """


class ResponseReader(object):  # pylint: disable=useless-object-inheritance
    """
        Helper class for extracting logging-relevant information from HTTP response object
    """

    def get_status_code(self, response):
        """
        get response's integer status code

        :param response:
        """
        raise NotImplementedError

    def get_response_size(self, response):
        """
        get response's size in bytes

        :param response:
        """
        raise NotImplementedError

    # pylint: disable=no-self-use
    def get_content_type(self, response):
        """
        get response's MIME/media type

        :param response:
        """
        return response.content_type



""" Sanic request reader """
from sap.cf_logging import defaults
from sap.cf_logging.core.request_reader import RequestReader


class SanicRequestReader(RequestReader):
    """ Reads data from Sanic request """

    def get_remote_user(self, request):
        return defaults.UNKNOWN

    def get_protocol(self, request):
        return defaults.UNKNOWN

    def get_content_length(self, request):
        return defaults.UNKNOWN

    def get_remote_ip(self, request):
        return request.ip[0]

    def get_remote_port(self, request):
        return None



""" Flask request reader """
from sap.cf_logging import defaults
from sap.cf_logging.core.request_reader import RequestReader


class FlaskRequestReader(RequestReader):
    """ Reads data from Flask request """

    def get_remote_user(self, request):
        if request.authorization is not None:
            return request.authorization.username

        return defaults.UNKNOWN

    def get_protocol(self, request):
        return request.environ.get('SERVER_PROTOCOL')

    def get_content_length(self, request):
        return request.content_length

    def get_remote_ip(self, request):
        return request.remote_addr

    def get_remote_port(self, request):
        return request.environ.get('REMOTE_PORT')



""" Module framework """
import sys
from sap.cf_logging.core.context import Context
from sap.cf_logging.core.request_reader import RequestReader
from sap.cf_logging.core.response_reader import ResponseReader


STR_CLASS = str if sys.version_info[0] == 3 else basestring # pylint: disable=undefined-variable


def _check_instance(obj, clazz):
    if not isinstance(obj, clazz):
        raise TypeError('Provided object is not valid {}'.format(clazz.__name__))


class Framework(object):  # pylint: disable=useless-object-inheritance
    """ Framework class holds Context, RequestReader, ResponseReader """

    # pylint: disable=too-many-arguments
    def __init__(self, name, context, request_reader, response_reader, custom_fields=None):
        if not name or not isinstance(name, STR_CLASS):
            raise TypeError('Provided name is not valid string')
        _check_instance(context, Context)
        _check_instance(request_reader, RequestReader)
        _check_instance(response_reader, ResponseReader)
        self._name = name
        self._context = context
        self._request_reader = request_reader
        self._response_reader = response_reader
        self._custom_fields = custom_fields or {}

    @property
    def custom_fields(self):
        """ Get the custom fields """
        return self._custom_fields

    @property
    def context(self):
        """ Get Context """
        return self._context

    @property
    def request_reader(self):
        """ Get RequestReader """
        return self._request_reader

    @property
    def response_reader(self):
        """ Get Response Reader """
        return self._response_reader



""" Flask logging context - used by the logging package to keep
request specific data, needed for logging purposes.
For example correlation_id needs to be stored during request processing,
so all log entries contain it.
"""
from flask import g
from sap.cf_logging.core.context import Context


class FlaskContext(Context):
    """ Stores logging context in Flask's request scope """

    def set(self, key, value, request):
        if g:
            setattr(g, key, value)

    def get(self, key, request):
        return getattr(g, key, None) if g else None



""" Logging module constants """
REQUEST_KEY = 'request'
RESPONSE_KEY = 'response'

LOG_SENSITIVE_CONNECTION_DATA = 'LOG_SENSITIVE_CONNECTION_DATA'
LOG_REMOTE_USER = 'LOG_REMOTE_USER'
LOG_REFERER = 'LOG_REFERER'

STACKTRACE_MAX_SIZE = 55 * 1024



""" Falcon request reader """

from sap.cf_logging import defaults
from sap.cf_logging.core.request_reader import RequestReader


class FalconRequestReader(RequestReader):
    """ Read log related properties out of falcon request """

    def __init__(self, username_key):
        self._username_key = username_key

    def get_remote_user(self, request):
        user = request.context.get('user')
        if user and self._username_key:
            return user.get(self._username_key) or defaults.UNKNOWN

        return defaults.UNKNOWN

    def get_protocol(self, request):
        return request.scheme

    def get_content_length(self, request):
        # pylint: disable=duplicate-code
        return request.content_length

    def get_remote_ip(self, request):
        # pylint: disable=duplicate-code
        return request.remote_addr

    def get_remote_port(self, request):
        return defaults.UNKNOWN



""" Logging support for Django based applications """
import logging
from datetime import datetime

from sap import cf_logging
from sap.cf_logging import defaults
from sap.cf_logging.core.constants import REQUEST_KEY, RESPONSE_KEY
from sap.cf_logging.core.framework import Framework
from sap.cf_logging.django_logging.context import DjangoContext
from sap.cf_logging.django_logging.request_reader import DjangoRequestReader
from sap.cf_logging.django_logging.response_reader import DjangoResponseReader

DJANGO_FRAMEWORK_NAME = 'django.framework'


class LoggingMiddleware(object):  # pylint: disable=useless-object-inheritance
    """ Django logging middleware """

    def __init__(self, get_response, logger_name='cf.django.logger'):
        self._logger_name = logger_name
        self._get_response = get_response

    def __call__(self, request):
        self.process_request(request)
        response = self._get_response(request)

        response = self.process_response(request, response)
        return response

    def process_request(self, request):# pylint: disable=no-self-use
        """
        Process the request before routing it.

        :param request: - Django Request object
        """
        framework = cf_logging.FRAMEWORK
        cid = framework.request_reader.get_correlation_id(request)
        framework.context.set_correlation_id(cid, request)
        framework.context.set('request_started_at', datetime.utcnow(), request)

    def process_response(self, request, response):
        """
        Post-processing of the response (after routing).

        :param request: - Django Request object
        :param request: - Django Response object
        """
        cf_logging.FRAMEWORK.context.set(
            'response_sent_at', datetime.utcnow(), request)
        extra = {REQUEST_KEY: request, RESPONSE_KEY: response}
        logging.getLogger(self._logger_name).info('', extra=extra)
        return response


def init(level=defaults.DEFAULT_LOGGING_LEVEL, custom_fields=None):
    """
    Initializes logging in JSON format.

    :param level: - valid log level from standard logging package (optional)
    """
    framework = Framework(DJANGO_FRAMEWORK_NAME, DjangoContext(),
                          DjangoRequestReader(), DjangoResponseReader(),
                          custom_fields=custom_fields)

    cf_logging.init(framework, level)



""" Flask response reader """
from sap.cf_logging.core.response_reader import ResponseReader


class FlaskResponseReader(ResponseReader):
    """ Implements Flask specific `ResponseReader` """

    def get_status_code(self, response):
        return response.status_code

    def get_response_size(self, response):
        return response.calculate_content_length()



""" Job logging context - used by the logging package to keep log data """
import threading
from sap.cf_logging.core.context import Context


class JobContext(Context, threading.local):
    """ Stores logging context in dict """

    def __init__(self):
        super(JobContext, self).__init__()
        self._mem_store = {}

    def set(self, key, value, request):
        self._mem_store[key] = value

    def get(self, key, request):
        return self._mem_store[key] if key in self._mem_store else None



""" Django response reader """
from sap.cf_logging.core.response_reader import ResponseReader


class DjangoResponseReader(ResponseReader):
    """ Read log related properties out of Django response """

    def get_status_code(self, response):
        return response.status_code

    def get_response_size(self, response):
        return len(response.content)

    def get_content_type(self, response):
        return response.get('Content-Type')



""" Module for the JsonFormatter """
import json
import logging
import sys
from sap.cf_logging.record.simple_log_record import SimpleLogRecord

def _default_serializer(obj):
    return str(obj)

if sys.version_info[0] == 3:
    def _encode(obj):
        return json.dumps(obj, default=_default_serializer)
else:
    def _encode(obj):
        return unicode(json.dumps(obj, default=_default_serializer))  # pylint: disable=undefined-variable


class JsonFormatter(logging.Formatter):
    """
    Format application log in JSON format
    """

    def format(self, record):
        """ Format the known log records in JSON format """
        if isinstance(record, SimpleLogRecord):
            return _encode(record.format())
        return super(JsonFormatter, self).format(record)



""" Framework to be used by worker/job running applications """
from sap.cf_logging.core.framework import Framework
from sap.cf_logging.job_logging.context import JobContext
from sap.cf_logging.core.request_reader import RequestReader
from sap.cf_logging.core.response_reader import ResponseReader

JOB_FRAMEWORK_NAME = 'job.framework'


class JobFramework(Framework):
    """ Simple framework using default request and response readers.
    Uses JobContext to keeping properties in memory  """

    def __init__(self, context=None, custom_fields=None):
        super(JobFramework, self).__init__(
            JOB_FRAMEWORK_NAME,
            context or JobContext(),
            RequestReader(),
            ResponseReader(),
            custom_fields=custom_fields
        )



""" Module that contains the CfLogger class """
import logging
import sys

from sap.cf_logging import defaults
from sap.cf_logging.core.constants import REQUEST_KEY, RESPONSE_KEY
from sap.cf_logging.core.framework import Framework
from sap.cf_logging.formatters.json_formatter import JsonFormatter
from sap.cf_logging.job_logging.framework import JobFramework
from sap.cf_logging.record.request_log_record import RequestWebRecord
from sap.cf_logging.record.simple_log_record import SimpleLogRecord

__version__ = '4.2.3'

_SETUP_DONE = False
FRAMEWORK = None


class CfLogger(logging.Logger):
    """ CfLogger class inherits from logging.Logger and makes custom
        log messages
    """

    # pylint: disable=too-many-arguments,arguments-differ,keyword-arg-before-vararg
    def makeRecord(self, name, level, fn, lno, msg, msgargs, exc_info,
                   func=None, extra=None, *args, **kwargs):
        """ Returns SimpleLogMessage or a RequestWebRecord depending on the extra variable """
        # check what record type this is
        cls = None
        if extra is not None and REQUEST_KEY in extra and RESPONSE_KEY in extra:
            cls = RequestWebRecord
        else:
            cls = SimpleLogRecord

        return cls(extra, FRAMEWORK, name, level, fn, lno, msg, msgargs, exc_info,
                   func, *args, **kwargs)

def init(cfl_framework=None, level=defaults.DEFAULT_LOGGING_LEVEL, custom_fields=None):
    """ Initialize function. It sets up the logging library to output JSON
        formatted messages.

        Optional arguments framework to use and logging.level
    """
    global FRAMEWORK  # pylint: disable=global-statement
    global _SETUP_DONE  # pylint: disable=global-statement
    if _SETUP_DONE:
        raise RuntimeError('cf_logging already initialized')

    if cfl_framework is not None and not isinstance(cfl_framework, Framework):
        raise TypeError('expecting framework of type {}'.format(Framework.__name__))

    _SETUP_DONE = True
    FRAMEWORK = cfl_framework or JobFramework(custom_fields=custom_fields)

    logging.setLoggerClass(CfLogger)

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(JsonFormatter())

    root = logging.getLogger()
    root.setLevel(level)
    root.addHandler(handler)



""" request_reader provides an interface for the RequestReader class """
import uuid

CORRELATION_ID_HEADERS = ['X-Correlation-ID',
                          'X-CorrelationID', 'X-Request-ID', 'X-Vcap-Request-Id']


class RequestReader(object):  # pylint: disable=useless-object-inheritance
    """
        Helper class for extracting logging-relevant information from HTTP request object
    """

    def get_correlation_id(self, request):
        """
        We assume that the request is a valid request object.

        :param request:
        """
        if request is None:
            return None

        for header in CORRELATION_ID_HEADERS:
            value = self.get_http_header(request, header)
            if value is not None:
                return value
        return str(uuid.uuid1())

    # pylint: disable=no-self-use
    def get_http_header(self, request, header_name, default=None):
        """
        get the HTTP header's value given its name

        :param request:
        :param header_name: name of header
        :param default: default value if header is not found
        :return:
        """
        if request is None or not hasattr(request, 'headers'):
            return default

        if header_name in request.headers:
            return request.headers.get(header_name)
        if header_name.upper() in request.headers:
            return request.headers.get(header_name.upper())
        return default

    def get_remote_user(self, request):
        """

        :param request:
        """
        raise NotImplementedError

    def get_protocol(self, request):
        """
        We assume that request is a valid request object.
        Gets the request protocol (e.g. HTTP/1.1).

        :return: The request protocol or None if it cannot be determined
        """
        raise NotImplementedError

    # pylint: disable=no-self-use
    def get_path(self, request):
        """
        We assume that request is a valid request object.
        Gets the request path.

       :return: the request path (e.g. /index.html)
        """
        return request.path

    def get_content_length(self, request):
        """
        We assume that request is a valid request object.
        The content length of the request.

        :return: the content length of the request or None if it cannot be determined
        """
        raise NotImplementedError

    # pylint: disable=no-self-use
    def get_method(self, request):
        """
        We assume that request is a valid request object.
        Gets the request method (e.g. GET, POST, etc.).

        :return: The request method or None if it cannot be determined
        """
        return request.method

    def get_remote_ip(self, request):
        """
        We assume that request is a valid request object.
        Gets the remote IP of the request initiator.

        :return: An ip address or None if it cannot be determined
        """
        raise NotImplementedError

    def get_remote_port(self, request):
        """
        We assume that request is a valid request object.
        Gets the remote port of the request initiator.

        :return: A port or None if it cannot be determined
        """
        raise NotImplementedError



""" Logging support for Falcon https://falconframework.org/ based applications """
import logging
from datetime import datetime

import falcon

from sap import cf_logging
from sap.cf_logging import defaults
from sap.cf_logging.core.constants import REQUEST_KEY, RESPONSE_KEY
from sap.cf_logging.core.framework import Framework
from sap.cf_logging.falcon_logging.context import FalconContext
from sap.cf_logging.falcon_logging.request_reader import FalconRequestReader
from sap.cf_logging.falcon_logging.response_reader import FalconResponseReader

FALCON_FRAMEWORK_NAME = 'falcon.framework'


class LoggingMiddleware(object):  # pylint: disable=useless-object-inheritance
    """ Falcon logging middleware """

    def __init__(self, logger_name='cf.falcon.logger'):
        self._logger_name = logger_name

    def process_request(self, request, response): # pylint: disable=unused-argument,no-self-use
        """Process the request before routing it.

        :param request: - Falcon Request object
        :param response: - Falcon Response object
        """
        framework = cf_logging.FRAMEWORK
        cid = framework.request_reader.get_correlation_id(request)
        framework.context.set_correlation_id(cid, request)
        framework.context.set('request_started_at', datetime.utcnow(), request)

    def process_response(self, request, response, resource, req_succeeded): # pylint: disable=unused-argument
        """Post-processing of the response (after routing).

        :param request: - Falcon Request object
        :param response: - Falcon Response object
        :param resource: - Falcon Resource object to which the request was routed
        :param req_succeeded: - True if no exceptions were raised while
            the framework processed and routed the request
        """
        cf_logging.FRAMEWORK.context.set(
            'response_sent_at', datetime.utcnow(), request)
        extra = {REQUEST_KEY: request, RESPONSE_KEY: response}
        logging.getLogger(self._logger_name).info('', extra=extra)


def init(app, level=defaults.DEFAULT_LOGGING_LEVEL, username_key='username', custom_fields=None):
    """ Initializes logging in JSON format.

    :param app: - Falcon application object
    :param level: - valid log level from standard logging package (optional)
    :param username_key: key used by the framework to get the username
        out of the request user, set in the request context,
        like `request.context.get('user').get(key)`
    """
    if not isinstance(app, falcon.API):
        raise TypeError('application should be instance of Falcon API')

    framework = Framework(FALCON_FRAMEWORK_NAME, FalconContext(),
                          FalconRequestReader(username_key), FalconResponseReader(),
                          custom_fields=custom_fields)
    cf_logging.init(framework, level)



""" Sanic response reader """
from sap.cf_logging.core.response_reader import ResponseReader

CONTENT_LENGTH = 'Content-Length'


class SanicResponseReader(ResponseReader):
    """ Implements Sanic specific `ResponseReader` """

    def get_status_code(self, response):
        return response.status

    def get_response_size(self, response):
        if CONTENT_LENGTH in response.headers:
            return response.headers.get(CONTENT_LENGTH)
        return None



""" Util module with helper functions """
import json
import os

from sap.cf_logging.defaults import UNIX_EPOCH

VCAP_APPLICATION = json.loads(os.getenv('VCAP_APPLICATION', default='{}'))


def get_vcap_param(param_name, default=None):
    """
    Returns a parameter from the VCAP_APPLICATION environment variable
    :param param_name: the name of the parameter
    :param default: a default value if it cannot be found
    :return: The value of the parameter, the default if it cannot be found,
    or '-' if the default is not specified.
    """
    return VCAP_APPLICATION.get(param_name, default)


def epoch_nano_second(datetime_):
    """ Returns the nanoseconds since epoch time """
    return int((datetime_ - UNIX_EPOCH).total_seconds()) * 1000000000 + datetime_.microsecond * 1000


def iso_time_format(datetime_):
    """ Returns ISO time formatted string """
    return '%04d-%02d-%02dT%02d:%02d:%02d.%03dZ' % (
        datetime_.year, datetime_.month, datetime_.day, datetime_.hour,
        datetime_.minute, datetime_.second, int(datetime_.microsecond / 1000))


def time_delta_ms(start, end):
    """ Returns the delta time between to datetime objects """
    time_delta = end - start
    return int(time_delta.total_seconds()) * 1000 + \
           int(time_delta.microseconds / 1000)


def parse_int(_int, default):
    """ Parses an int and returns the result
        A default can be provided in case the parse fails
    """
    try:
        integer = int(_int)
    except:  # pylint: disable-msg=bare-except
        integer = default
    return integer



""" Default values used inside the framework """
import logging

from datetime import datetime


UNKNOWN = '-'
REDACTED = 'redacted'
RESPONSE_SIZE_B = -1
RESPONSE_TIME_MS = 0
REQUEST_SIZE_B = -1
STATUS = -1

DEFAULT_LOGGING_LEVEL = logging.INFO

UNIX_EPOCH = datetime(1970, 1, 1)



""" Module that holds the RequestWebRecord class """
import os
from sap.cf_logging import defaults
from sap.cf_logging.core.constants import REQUEST_KEY, RESPONSE_KEY, \
    LOG_SENSITIVE_CONNECTION_DATA, LOG_REMOTE_USER, LOG_REFERER
from sap.cf_logging.record import util
from sap.cf_logging.record.simple_log_record import SimpleLogRecord


PROPS = ['type', 'direction', 'remote_user', 'request', 'referer',
         'x_forwarded_for', 'protocol', 'method', 'remote_ip',
         'request_size_b', 'remote_host', 'remote_port', 'request_received_at',
         'response_sent_at', 'response_time_ms', 'response_status',
         'response_size_b', 'response_content_type']

# pylint: disable=too-many-instance-attributes


class RequestWebRecord(SimpleLogRecord):
    """ RequestWebRecord class holds request/response log data
        It is used in the formatter class
    """

    # pylint: disable=too-many-locals
    def __init__(self, extra, framework, *args, **kwargs):

        super(RequestWebRecord, self).__init__(extra, framework, *args, **kwargs)

        context = framework.context
        request_reader = framework.request_reader
        response_reader = framework.response_reader

        request = extra[REQUEST_KEY]
        response = extra[RESPONSE_KEY]

        props = dict((key, value) for key, value in extra.items()
                     if key not in [RESPONSE_KEY, REQUEST_KEY])
        for key, value in props.items():
            setattr(self, key, value)

        length = request_reader.get_content_length(request)
        remote_ip = request_reader.get_remote_ip(request)

        self.type = 'request'
        self.direction = 'IN'
        self.remote_user = request_reader.get_remote_user(request)
        self.request = request_reader.get_path(request)
        self.referer = request_reader.get_http_header(
            request, 'referer', defaults.UNKNOWN)
        self.x_forwarded_for = request_reader.get_http_header(
            request, 'x-forwarded-for', defaults.UNKNOWN)
        self.protocol = request_reader.get_protocol(request)
        self.method = request_reader.get_method(request)
        self.remote_ip = remote_ip
        self.request_size_b = util.parse_int(length, -1)
        self.remote_host = remote_ip
        self.remote_port = request_reader.get_remote_port(
            request) or defaults.UNKNOWN

        request_start = context.get(
            'request_started_at', request) or defaults.UNIX_EPOCH
        self.request_received_at = util.iso_time_format(request_start)

        # response related
        response_sent_at = context.get(
            'response_sent_at', request) or defaults.UNIX_EPOCH
        self.response_sent_at = util.iso_time_format(response_sent_at)
        self.response_time_ms = util.time_delta_ms(
            request_start, response_sent_at)

        self.response_status = util.parse_int(
            response_reader.get_status_code(response), defaults.STATUS)
        self.response_size_b = util.parse_int(
            response_reader.get_response_size(response), defaults.RESPONSE_SIZE_B)
        self.response_content_type = response_reader.get_content_type(response)

        self._hide_sensitive_fields()

    def format(self):
        record = super(RequestWebRecord, self).format_cf_attributes()
        request_properties = dict(
            (key, value) for key, value in self.__dict__.items() if key in PROPS)
        record.update(request_properties)
        return record

    def _hide_sensitive_fields(self):
        if os.environ.get(LOG_SENSITIVE_CONNECTION_DATA, 'false').lower() != 'true':
            self.remote_ip = defaults.REDACTED
            self.remote_host = defaults.REDACTED
            self.remote_port = defaults.REDACTED
            self.x_forwarded_for = defaults.REDACTED

        if os.environ.get(LOG_REMOTE_USER, 'false').lower() != 'true':
            self.remote_user = defaults.REDACTED

        if os.environ.get(LOG_REFERER, 'false').lower() != 'true':
            self.referer = defaults.REDACTED



"""
Django logging context - used by the logging package to keep
request specific data, needed for logging purposes.
For example correlation_id needs to be stored during request processing,
so all log entries contain it.
"""

from sap.cf_logging.core.context import Context


def _init_context(request):
    if not hasattr(request, 'context'):
        request.context = {}

class DjangoContext(Context):
    """ Stores logging context in Django's request objecct """

    def set(self, key, value, request):
        if request is None:
            return
        _init_context(request)
        request.context[key] = value

    def get(self, key, request):
        if request is None:
            return None
        _init_context(request)
        return request.context.get(key) if request else None



""" Flask logging support package
- Configures standard logging package to produce JSON
- Produces info request log entry per request
"""
import logging
from datetime import datetime
from functools import wraps
from sanic import Sanic

from sap import cf_logging
from sap.cf_logging import defaults
from sap.cf_logging.core.constants import REQUEST_KEY, RESPONSE_KEY
from sap.cf_logging.core.framework import Framework
from sap.cf_logging.sanic_logging.context import SanicContext
from sap.cf_logging.sanic_logging.request_reader import SanicRequestReader
from sap.cf_logging.sanic_logging.response_reader import SanicResponseReader

SANIC_FRAMEWORK_NAME = 'sanic.framework'


def before_request(wrapped):
    """ Use as decorator on Sanic's before_request handler
    Handles correlation_id by setting it in the context for log records
    """

    @wraps(wrapped)
    def _wrapper(request):
        correlation_id = cf_logging.FRAMEWORK.request_reader.get_correlation_id(request)
        cf_logging.FRAMEWORK.context.set_correlation_id(correlation_id, request)
        cf_logging.FRAMEWORK.context.set('request_started_at', datetime.utcnow(), request)
        return wrapped(request)

    return _wrapper


def after_request(wrapped):
    """ Use as decorator on Sanic after_request handler
    Creates info log record per request
    """

    @wraps(wrapped)
    def _wrapper(request, response):
        cf_logging.FRAMEWORK.context.set('response_sent_at', datetime.utcnow(), request)
        extra = {REQUEST_KEY: request, RESPONSE_KEY: response}
        logging.getLogger('cf.sanic.logger').info('', extra=extra)
        return wrapped(request, response)

    return _wrapper


def init(app, level=defaults.DEFAULT_LOGGING_LEVEL, custom_framework=None, custom_fields=None):
    """ Initializes logging in JSON format.

    Adds before and after request handlers to the `app` object to enable request info log.
    :param app: - Flask application object
    :param level: - valid log level from standard logging package (optional)
    :param custom_framework: - `Framework` instance - use in case you need
        to change request processing behaviour for example to customize context storage
    """
    if not isinstance(app, Sanic):
        raise TypeError('application should be instance of Sanic')

    framework = custom_framework or \
                Framework(
                    SANIC_FRAMEWORK_NAME,
                    SanicContext(),
                    SanicRequestReader(),
                    SanicResponseReader(),
                    custom_fields=custom_fields
                )

    cf_logging.init(framework, level)

    @app.middleware('request')
    @before_request
    def _before_request(request):  # pylint: disable=unused-argument
        pass

    @app.middleware('response')
    @after_request
    def _after_request(request, response):  # pylint: disable=unused-argument
        pass



""" Module SimpleLogRecord """
import logging
import traceback

from datetime import datetime
from sap.cf_logging import defaults
from sap.cf_logging.core.constants import REQUEST_KEY, RESPONSE_KEY
from sap.cf_logging.record import application_info
from sap.cf_logging.record import util

from sap.cf_logging.formatters.stacktrace_formatter import format_stacktrace

_SKIP_ATTRIBUTES = ["type", "written_at", "written_ts", "correlation_id", "remote_user", "referer",
                    "x_forwarded_for", "protocol", "method", "remote_ip", "request_size_b",
                    "remote_host", "remote_port", "request_received_at", "direction",
                    "response_time_ms", "response_status", "response_size_b",
                    "response_content_type", "response_sent_at", REQUEST_KEY, RESPONSE_KEY]


class SimpleLogRecord(logging.LogRecord):
    """ SimpleLogRecord class holds data for user logged messages """

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, extra, framework, *args, **kwargs):
        super(SimpleLogRecord, self).__init__(*args, **kwargs)

        utcnow = datetime.utcnow()
        self.written_at = util.iso_time_format(utcnow)
        self.written_ts = util.epoch_nano_second(utcnow)

        request = extra[REQUEST_KEY] if extra and REQUEST_KEY in extra else None

        self.correlation_id = framework.context.get_correlation_id(request) or defaults.UNKNOWN

        self.custom_fields = {}
        for key, value in framework.custom_fields.items():
            if extra and key in extra:
                if extra[key] is not None:
                    self.custom_fields[key] = extra[key]
            elif value is not None:
                self.custom_fields[key] = value

        self.extra = dict((key, value) for key, value in extra.items()
                          if key not in _SKIP_ATTRIBUTES and
                          key not in framework.custom_fields.keys()) if extra else {}
        for key, value in self.extra.items():
            setattr(self, key, value)

    def format_cf_attributes(self):
        """ Add common and Cloud Foundry environment specific attributes """
        record = {
            'component_id': application_info.COMPONENT_ID,
            'component_name': application_info.COMPONENT_NAME,
            'component_instance': application_info.COMPONENT_INSTANCE,
            'space_id': application_info.SPACE_ID,
            'space_name': application_info.SPACE_NAME,
            'container_id': application_info.CONTAINER_ID,
            'component_type': application_info.COMPONENT_TYPE,
            'written_at': self.written_at,
            'written_ts': self.written_ts,
            'correlation_id': self.correlation_id,
            'layer': application_info.LAYER
        }
        return record

    def format(self):
        """ Returns a dict record with properties to be logged """
        record = self.format_cf_attributes()
        record.update({
            'type': 'log',
            'logger': self.name,
            'thread': self.threadName,
            'level': self.levelname,
            'msg': self.getMessage(),
        })

        if self.levelno == logging.ERROR and self.exc_info:
            stacktrace = ''.join(traceback.format_exception(*self.exc_info))
            stacktrace = format_stacktrace(stacktrace)
            record['stacktrace'] = stacktrace.split('\n')
            record['msg'] += "\n"
            record['msg'] += stacktrace


        record.update(self.extra)
        if len(self.custom_fields) > 0:
            record.update(self._format_custom_fields())
        return record

    def _format_custom_fields(self):
        res = {"#cf": {"string": []}}
        for i, (key, value) in enumerate(self.custom_fields.items()):
            res['#cf']['string'].append(
                {"k": str(key), "v": str(value), "i": i}
            )
        return res



''' setup.py '''
import ast
import re

from setuptools import find_packages
from setuptools import setup

_VERSION_REGEX = re.compile(r'__version__\s+=\s+(.*)')
with open('./sap/cf_logging/__init__.py', 'rb') as f:
    VERSION = str(ast.literal_eval(_VERSION_REGEX.search(f.read().decode('utf-8')).group(1)))

setup(
    name='sap_cf_logging',
    version=VERSION,
    url='https://github.com/SAP/cf-python-logging-support',
    license='Apache License, Version 2.0',
    author='SAP',
    description='Python logging library to emit JSON logs in a SAP CloudFoundry environment',
    long_description_content_type='text/x-rst',
    packages=find_packages(include=['sap*']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[],
    classifiers=[  # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development',
        'Topic :: System :: Logging',
    ]
)



""" Falcon logging context - used by the logging package to keep
request specific data, needed for logging purposes.
For example correlation_id needs to be stored during request processing,
so all log entries contain it.
"""

from sap.cf_logging.core.context import Context


class FalconContext(Context):
    """ Stores logging context in Falcon's request object"""

    def set(self, key, value, request):
        if request:
            request.context[key] = value

    def get(self, key, request):
        return request.context.get(key) if request else None



"""
Module for formatting utilities for stacktrace
generated by user logging.exception call
"""
import re

from sap.cf_logging.core import constants


def format_stacktrace(stacktrace):
    """
    Removes tab characters
    Truncates stacktrace to maximum size

    :param stacktrace: string representation of a stacktrace
    """
    if not isinstance(stacktrace, str):
        return ''

    stacktrace = re.sub('\t', '  ', stacktrace)

    if len(stacktrace) <= constants.STACKTRACE_MAX_SIZE:
        return stacktrace

    stacktrace_beginning = _stacktrace_beginning(
        stacktrace, constants.STACKTRACE_MAX_SIZE // 3
    )

    stacktrace_end = _stacktrace_end(
        stacktrace, (constants.STACKTRACE_MAX_SIZE // 3) * 2
    )

    new_stacktrace = "-------- STACK TRACE TRUNCATED --------\n" + stacktrace_beginning +\
        "-------- OMITTED --------\n" + stacktrace_end

    return new_stacktrace

def _stacktrace_beginning(stacktrace, size):
    """ Gets the first `size` bytes of the stacktrace """
    if len(stacktrace) <= size:
        return stacktrace

    return stacktrace[:size]

def _stacktrace_end(stacktrace, size):
    """ Gets the last `size` bytes of the stacktrace """
    stacktrace_length = len(stacktrace)
    if stacktrace_length <= size:
        return stacktrace

    return stacktrace[:-(stacktrace_length-size)]



""" Module for record constants """
import os

from sap.cf_logging import defaults
from sap.cf_logging.record import util


LAYER = 'python'
COMPONENT_ID = util.get_vcap_param('application_id', defaults.UNKNOWN)
COMPONENT_NAME = util.get_vcap_param('name', defaults.UNKNOWN)
COMPONENT_INSTANCE = int(os.getenv('CF_INSTANCE_INDEX', "0"))
SPACE_ID = util.get_vcap_param('space_id', defaults.UNKNOWN)
SPACE_NAME = util.get_vcap_param('space_name', defaults.UNKNOWN)
CONTAINER_ID = os.getenv('CF_INSTANCE_IP', defaults.UNKNOWN)
COMPONENT_TYPE = 'application'



""" Flask logging support package
- Configures standard logging package to produce JSON
- Produces info request log entry per request
"""
import logging
from datetime import datetime
from functools import wraps
import flask
from flask import request

from sap import cf_logging
from sap.cf_logging import defaults
from sap.cf_logging.core.constants import REQUEST_KEY, RESPONSE_KEY
from sap.cf_logging.core.framework import Framework
from sap.cf_logging.flask_logging.context import FlaskContext
from sap.cf_logging.flask_logging.request_reader import FlaskRequestReader
from sap.cf_logging.flask_logging.response_reader import FlaskResponseReader

FLASK_FRAMEWORK_NAME = 'flask.framework'


def before_request(wrapped):
    """ Use as a decorator on Flask before_request handler
    Handles correlation_id by setting it in the context for log records
    """
    @wraps(wrapped)
    def _wrapper():
        framework = cf_logging.FRAMEWORK
        cid = framework.request_reader.get_correlation_id(request)
        framework.context.set_correlation_id(cid, request)
        framework.context.set('request_started_at', datetime.utcnow(), request)
        return wrapped()
    return _wrapper


def after_request(wrapped):
    """ Use as a decorator on Flask after_request handler
    Creates info log record per request
    """
    @wraps(wrapped)
    def _wrapper(response):
        cf_logging.FRAMEWORK.context.set(
            'response_sent_at', datetime.utcnow(), request)
        extra = {REQUEST_KEY: request, RESPONSE_KEY: response}
        logging.getLogger('cf.flask.logger').info('', extra=extra)
        return wrapped(response)
    return _wrapper


def init(app, level=defaults.DEFAULT_LOGGING_LEVEL, custom_fields=None):
    """ Initializes logging in JSON format.

    Adds before and after request handlers to `app` object to enable request info log.
    :param app: - Flask application object
    :param level: - valid log level from standard logging package (optional)
    """
    if not isinstance(app, flask.Flask):
        raise TypeError('application should be instance of Flask')

    _init_framework(level, custom_fields=custom_fields)

    @app.before_request
    @before_request
    def _app_before_request():
        pass

    @app.after_request
    @after_request
    def _app_after_request(response):
        return response


def _init_framework(level, custom_fields):
    logging.getLogger('werkzeug').disabled = True

    framework = Framework(FLASK_FRAMEWORK_NAME,
                          FlaskContext(), FlaskRequestReader(), FlaskResponseReader(),
                          custom_fields=custom_fields)
    cf_logging.init(framework, level)
